"""Command line interface utilities for the FreeWili library.

This module provides a command line interface to find and control FreeWili boards.
"""

import sys

from result import Err, Ok, Result

from freewili.fw import FreeWili


def exit_with_error(msg: str, exit_code: int = 1) -> None:
    """A function that prints an error message to the stderr and exits the program with a specified exit code.

    Parameters:
    -----------
        msg: str
            The error message to be printed.
        exit_code: int, optional
            The exit code to be used when exiting the program, defaults to 1.

    Returns:
    --------
        None
    """
    print(msg, file=sys.stderr)
    sys.exit(exit_code)


def get_device(
    index: int,
    devices: tuple[FreeWili, ...],
) -> Result[FreeWili, str]:
    """Get a FreeWili by index.

    Parameters:
    -----------
        index: int
            The index to be checked.

        devices: tuple[FreeWili]
            container of FreeWili to reference

    Returns:
    --------
        Result[FreeWili, str]:
            The FreeWili if the index is valid, otherwise an error message.
    """
    if index >= len(devices):
        return Err(f"Index {index} is out of range. There are only {len(devices)} devices.")
    return Ok(devices[index])



"""Tests for freewili.image module."""

import pathlib
import struct
import tempfile
from unittest.mock import MagicMock, mock_open, patch

from freewili.image import convert


class TestConvert:
    """Test cases for the convert function."""

    def test_convert_success_no_transparency(self) -> None:
        """Test successful conversion of image without transparency."""
        # Create a mock image
        mock_image = MagicMock()
        mock_image.size = (2, 2)  # 2x2 image

        # Mock pixel data without transparency (RGB)
        mock_pix = {
            (0, 0): (255, 0, 0),  # Red pixel
            (1, 0): (0, 255, 0),  # Green pixel
            (0, 1): (0, 0, 255),  # Blue pixel
            (1, 1): (255, 255, 255),  # White pixel
        }
        mock_image.load.return_value = mock_pix

        with patch("freewili.image.Image.open", return_value=mock_image), patch("builtins.open", mock_open()):
            input_path = pathlib.Path("test.png")
            output_path = pathlib.Path("test.fwi")

            result = convert(input_path, output_path)

            assert result.is_ok()
            assert "converted to FreeWili image file" in result.unwrap()

    def test_convert_success_with_transparency(self) -> None:
        """Test successful conversion of image with transparency."""
        mock_image = MagicMock()
        mock_image.size = (1, 1)  # 1x1 image

        # Mock pixel data with transparency (RGBA)
        mock_pix = {
            (0, 0): (255, 0, 0, 128)  # Red pixel with transparency
        }
        mock_image.load.return_value = mock_pix

        with (
            patch("freewili.image.Image.open", return_value=mock_image),
            patch("builtins.open", mock_open()) as mock_file,
        ):
            input_path = pathlib.Path("test.png")
            output_path = pathlib.Path("test.fwi")

            result = convert(input_path, output_path)

            assert result.is_ok()
            mock_file.assert_called_once_with(output_path, "wb")

    def test_convert_transparent_pixel(self) -> None:
        """Test conversion of fully transparent pixel."""
        mock_image = MagicMock()
        mock_image.size = (1, 1)

        # Mock pixel data with full transparency
        mock_pix = {
            (0, 0): (255, 0, 0, 0)  # Fully transparent pixel
        }
        mock_image.load.return_value = mock_pix

        with (
            patch("freewili.image.Image.open", return_value=mock_image),
            patch("builtins.open", mock_open()),
        ):
            input_path = pathlib.Path("test.png")
            output_path = pathlib.Path("test.fwi")

            result = convert(input_path, output_path)

            assert result.is_ok()

    def test_convert_image_open_failure(self) -> None:
        """Test handling of image opening failure."""
        with patch("freewili.image.Image.open", side_effect=Exception("Cannot open image")):
            input_path = pathlib.Path("nonexistent.png")
            output_path = pathlib.Path("test.fwi")

            result = convert(input_path, output_path)

            assert result.is_err()
            error_msg = result.unwrap_err()
            assert "Fail to open png or jpg file" in error_msg
            assert "Cannot open image" in error_msg

    def test_convert_output_file_write_failure(self) -> None:
        """Test handling of output file write failure."""
        mock_image = MagicMock()
        mock_image.size = (1, 1)

        with (
            patch("freewili.image.Image.open", return_value=mock_image),
            patch("builtins.open", side_effect=Exception("Cannot write file")),
        ):
            input_path = pathlib.Path("test.png")
            output_path = pathlib.Path("readonly.fwi")

            result = convert(input_path, output_path)

            assert result.is_err()
            error_msg = result.unwrap_err()
            assert "Can't write the file" in error_msg
            assert "Cannot write file" in error_msg

    def test_convert_header_format(self) -> None:
        """Test that the FWI header is written correctly."""
        mock_image = MagicMock()
        mock_image.size = (2, 3)  # 2x3 image

        mock_pix = {
            (0, 0): (0, 0, 0),
            (1, 0): (0, 0, 0),
            (0, 1): (0, 0, 0),
            (1, 1): (0, 0, 0),
            (0, 2): (0, 0, 0),
            (1, 2): (0, 0, 0),
        }
        mock_image.load.return_value = mock_pix

        with patch("freewili.image.Image.open", return_value=mock_image):
            with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
                input_path = pathlib.Path("test.png")
                output_path = pathlib.Path(tmp_file.name)

                result = convert(input_path, output_path)
                assert result.is_ok()

                # Read back the header to verify format
                with open(output_path, "rb") as f:
                    # Header ID: "FW01IMG\0"
                    header_id = f.read(8)
                    assert header_id == b"FW01IMG\0"

                    # Flags (4 bytes)
                    flags = struct.unpack("<I", f.read(4))[0]
                    assert flags == 1

                    # Total pixel count (4 bytes)
                    pixel_count = struct.unpack("<I", f.read(4))[0]
                    assert pixel_count == 6  # 2 * 3

                    # Width (2 bytes)
                    width = struct.unpack("<h", f.read(2))[0]
                    assert width == 2

                    # Height (2 bytes)
                    height = struct.unpack("<h", f.read(2))[0]
                    assert height == 3

    def test_convert_color_conversion(self) -> None:
        """Test RGB to 16-bit color conversion."""
        mock_image = MagicMock()
        mock_image.size = (1, 1)

        # Test with pure red color (255, 0, 0)
        mock_pix = {(0, 0): (255, 0, 0)}
        mock_image.load.return_value = mock_pix

        with patch("freewili.image.Image.open", return_value=mock_image):
            with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
                input_path = pathlib.Path("test.png")
                output_path = pathlib.Path(tmp_file.name)

                result = convert(input_path, output_path)
                assert result.is_ok()

                # Read the pixel data (skip 24-byte header)
                with open(output_path, "rb") as f:
                    f.seek(24)  # Skip header
                    pixel_data = f.read(2)  # Read one 16-bit pixel
                    pixel_value = struct.unpack("<H", pixel_data)[0]

                    # Pure red should convert to a non-zero value
                    assert pixel_value > 0



# type: ignore
"""PySide6 GUI example to display FreeWili events in real-time."""

import sys
from typing import Any

from PySide6.QtCore import QObject, QTimer, Signal
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QPushButton,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from freewili import FreeWili
from freewili.framing import ResponseFrame
from freewili.types import AccelData, BatteryData, ButtonData, EventType, GPIOData, IRData


class EventSignals(QObject):
    """Signals for thread-safe GUI updates."""

    accel_updated = Signal(AccelData)
    button_updated = Signal(ButtonData)
    ir_updated = Signal(IRData)
    battery_updated = Signal(BatteryData)
    gpio_updated = Signal(GPIOData)
    raw_event = Signal(str)


class FreeWiliEventsGUI(QMainWindow):
    """Main window for FreeWili events display."""

    def __init__(self):
        super().__init__()
        self.freewili = None
        self.event_timer = QTimer()
        self.signals = EventSignals()

        # Connect signals
        self.signals.accel_updated.connect(self.update_accel_display)
        self.signals.button_updated.connect(self.update_button_display)
        self.signals.ir_updated.connect(self.update_ir_display)
        self.signals.battery_updated.connect(self.update_battery_display)
        self.signals.gpio_updated.connect(self.update_gpio_display)
        self.signals.raw_event.connect(self.update_raw_events)

        self.setup_ui()
        self.setup_freewili()

    def setup_ui(self):
        """Set up the user interface."""
        self.setWindowTitle("FreeWili Events Monitor")
        self.setGeometry(100, 100, 800, 600)

        central_widget = QWidget()
        self.setCentralWidget(central_widget)

        main_layout = QVBoxLayout(central_widget)

        # Control buttons
        control_layout = QHBoxLayout()

        self.connect_btn = QPushButton("Connect")
        self.connect_btn.clicked.connect(self.toggle_connection)
        control_layout.addWidget(self.connect_btn)

        # Event enable checkboxes
        self.accel_cb = QCheckBox("Accelerometer")
        self.button_cb = QCheckBox("Buttons")
        self.ir_cb = QCheckBox("IR")
        self.battery_cb = QCheckBox("Battery")
        self.gpio_cb = QCheckBox("GPIO")

        for cb in [self.accel_cb, self.button_cb, self.ir_cb, self.battery_cb, self.gpio_cb]:
            cb.stateChanged.connect(self.update_event_subscriptions)
            control_layout.addWidget(cb)

        control_layout.addStretch()
        main_layout.addLayout(control_layout)

        # Data display areas
        data_layout = QHBoxLayout()

        # Left column - Sensor data
        left_layout = QVBoxLayout()

        # Accelerometer group
        self.accel_group = self.create_accel_group()
        left_layout.addWidget(self.accel_group)

        # Battery group
        self.battery_group = self.create_battery_group()
        left_layout.addWidget(self.battery_group)

        data_layout.addLayout(left_layout)

        # Right column - Digital inputs
        right_layout = QVBoxLayout()

        # Button group
        self.button_group = self.create_button_group()
        right_layout.addWidget(self.button_group)

        # IR group
        self.ir_group = self.create_ir_group()
        right_layout.addWidget(self.ir_group)

        # GPIO group
        self.gpio_group = self.create_gpio_group()
        right_layout.addWidget(self.gpio_group)

        # Raw events log
        self.raw_events_group = self.create_raw_events_group()
        right_layout.addWidget(self.raw_events_group)

        data_layout.addLayout(right_layout)
        main_layout.addLayout(data_layout)

    def create_accel_group(self):
        """Create accelerometer data display group."""
        group = QGroupBox("Accelerometer")
        layout = QGridLayout(group)

        # Labels and value displays
        self.accel_labels = {}
        fields = [
            ("G-Force:", "g"),
            ("X:", "x"),
            ("Y:", "y"),
            ("Z:", "z"),
            ("Temp °C:", "temp_c"),
            ("Temp °F:", "temp_f"),
        ]

        for i, (label_text, key) in enumerate(fields):
            label = QLabel(label_text)
            value_label = QLabel("--")
            value_label.setStyleSheet("font-weight: bold; color: blue;")

            layout.addWidget(label, i, 0)
            layout.addWidget(value_label, i, 1)

            self.accel_labels[key] = value_label

        return group

    def create_battery_group(self):
        """Create battery data display group."""
        group = QGroupBox("Battery")
        layout = QGridLayout(group)

        self.battery_labels = {}
        fields = [
            ("VBUS:", "vbus"),
            ("VSYS:", "vsys"),
            ("VBATT:", "vbatt"),
            ("ICHG:", "ichg"),
            ("Charging:", "charging"),
            ("Complete:", "charge_complete"),
        ]

        for i, (label_text, key) in enumerate(fields):
            label = QLabel(label_text)
            value_label = QLabel("--")
            value_label.setStyleSheet("font-weight: bold; color: green;")

            layout.addWidget(label, i, 0)
            layout.addWidget(value_label, i, 1)

            self.battery_labels[key] = value_label

        return group

    def create_button_group(self):
        """Create button state display group."""
        group = QGroupBox("Buttons")
        layout = QGridLayout(group)

        self.button_labels = {}
        buttons = [("Gray:", "gray"), ("Yellow:", "yellow"), ("Green:", "green"), ("Blue:", "blue"), ("Red:", "red")]

        for i, (label_text, key) in enumerate(buttons):
            label = QLabel(label_text)
            value_label = QLabel("Released")
            value_label.setStyleSheet("font-weight: bold;")

            layout.addWidget(label, i, 0)
            layout.addWidget(value_label, i, 1)

            self.button_labels[key] = value_label

        return group

    def create_ir_group(self):
        """Create IR data display group."""
        group = QGroupBox("IR Data")
        layout = QVBoxLayout(group)

        self.ir_value_label = QLabel("No data")
        self.ir_value_label.setStyleSheet("font-weight: bold; color: purple; font-family: monospace;")
        layout.addWidget(self.ir_value_label)

        return group

    def create_gpio_group(self):
        """Create GPIO state display group."""
        group = QGroupBox("GPIO States")
        layout = QGridLayout(group)

        self.gpio_labels = {}

        # Get GPIO mappings from types.py
        from freewili.types import GPIO_MAP

        # Create labels for each GPIO pin
        row = 0
        for _gpio_num, gpio_name in GPIO_MAP.items():
            # Show just the GPIO number and main function
            display_name = gpio_name.replace("/", " / ")

            label = QLabel(f"{display_name}:")
            value_label = QLabel("--")
            value_label.setStyleSheet("font-weight: bold; font-family: monospace;")

            layout.addWidget(label, row, 0)
            layout.addWidget(value_label, row, 1)

            self.gpio_labels[gpio_name] = value_label
            row += 1

        return group

    def create_raw_events_group(self):
        """Create raw events log group."""
        group = QGroupBox("Raw Events Log")
        layout = QVBoxLayout(group)

        self.raw_events_text = QTextEdit()
        # self.raw_events_text.setMaximumHeight(150)
        self.raw_events_text.setStyleSheet("font-family: monospace; font-size: 10px;")
        layout.addWidget(self.raw_events_text)

        clear_btn = QPushButton("Clear Log")
        clear_btn.clicked.connect(self.raw_events_text.clear)
        layout.addWidget(clear_btn)

        return group

    def setup_freewili(self):
        """Set up FreeWili connection and event processing."""
        self.event_timer.timeout.connect(self.process_events)
        self.event_timer.setInterval(50)  # Process events every 50ms

    def event_handler(self, event_type: EventType, frame: ResponseFrame, data: Any) -> None:
        """Handle events from FreeWili (runs in background thread)."""
        # Convert the data to the appropriate type and emit signals for thread-safe GUI updates
        try:
            if event_type == EventType.Accel and isinstance(data, AccelData):
                self.signals.accel_updated.emit(data)
            elif event_type == EventType.Button and isinstance(data, ButtonData):
                self.signals.button_updated.emit(data)
            elif event_type == EventType.IR and isinstance(data, IRData):
                self.signals.ir_updated.emit(data)
            elif event_type == EventType.Battery and isinstance(data, BatteryData):
                self.signals.battery_updated.emit(data)
            elif event_type == EventType.GPIO and isinstance(data, GPIOData):
                self.signals.gpio_updated.emit(data)

            # Also log to raw events
            self.signals.raw_event.emit(f"{event_type.name}: {data}")

        except Exception as e:
            self.signals.raw_event.emit(f"Error processing {event_type.name}: {e}")

    def update_accel_display(self, data: AccelData):
        """Update accelerometer display."""
        self.accel_labels["g"].setText(f"{data.g:.2f}g")
        self.accel_labels["x"].setText(f"{data.x:.1f}")
        self.accel_labels["y"].setText(f"{data.y:.1f}")
        self.accel_labels["z"].setText(f"{data.z:.1f}")
        self.accel_labels["temp_c"].setText(f"{data.temp_c:.1f}°C")
        self.accel_labels["temp_f"].setText(f"{data.temp_f:.1f}°F")

    def update_button_display(self, data: ButtonData):
        """Update button display."""
        buttons = {"gray": data.gray, "yellow": data.yellow, "green": data.green, "blue": data.blue, "red": data.red}

        for key, pressed in buttons.items():
            label = self.button_labels[key]
            if pressed:
                label.setText("PRESSED")
                label.setStyleSheet("font-weight: bold; color: green;")
            else:
                label.setText("Released")
                label.setStyleSheet("font-weight: bold; color: red;")

    def update_ir_display(self, data: IRData):
        """Update IR display."""
        hex_value = data.value.hex().upper()
        self.ir_value_label.setText(f"0x{hex_value}" if hex_value else "No data")

    def update_battery_display(self, data: BatteryData):
        """Update battery display."""
        self.battery_labels["vbus"].setText(f"{data.vbus:.1f}mV")
        self.battery_labels["vsys"].setText(f"{data.vsys:.1f}mV")
        self.battery_labels["vbatt"].setText(f"{data.vbatt:.1f}mV")
        self.battery_labels["ichg"].setText(f"{data.ichg}mA")
        self.battery_labels["charging"].setText("YES" if data.charging else "NO")
        self.battery_labels["charge_complete"].setText("YES" if data.charge_complete else "NO")

    def update_gpio_display(self, data: GPIOData):
        """Update GPIO display."""
        for gpio_name, value in data.pin.items():
            if gpio_name in self.gpio_labels:
                label = self.gpio_labels[gpio_name]
                if value == 1:
                    label.setText("HIGH")
                    label.setStyleSheet("font-weight: bold; color: red; font-family: monospace;")
                else:
                    label.setText("LOW")
                    label.setStyleSheet("font-weight: bold; color: blue; font-family: monospace;")

    def update_raw_events(self, event_text: str):
        """Update raw events log."""
        self.raw_events_text.append(event_text)

        # Keep log size manageable
        if self.raw_events_text.document().blockCount() > 100:
            cursor = self.raw_events_text.textCursor()
            cursor.movePosition(cursor.MoveOperation.Start)
            cursor.select(cursor.SelectionType.BlockUnderCursor)
            cursor.removeSelectedText()
            cursor.deleteChar()  # Remove the newline

    def toggle_connection(self):
        """Connect or disconnect from FreeWili."""
        if self.freewili is None:
            try:
                result = FreeWili.find_first()
                if result.is_err():
                    self.signals.raw_event.emit(f"Error: {result.unwrap_err()}")
                    return

                self.freewili = result.unwrap()
                self.freewili.set_event_callback(self.event_handler)

                # Open connection
                open_result = self.freewili.open()
                if open_result.is_err():
                    self.signals.raw_event.emit(f"Error opening connection: {open_result.unwrap_err()}")
                    self.freewili = None
                    return

                self.connect_btn.setText("Disconnect")
                self.event_timer.start()
                self.signals.raw_event.emit("Connected to FreeWili")

                # Enable default events
                self.accel_cb.setChecked(True)
                self.button_cb.setChecked(True)
                self.ir_cb.setChecked(True)
                self.battery_cb.setChecked(True)
                self.gpio_cb.setChecked(True)

            except Exception as e:
                self.signals.raw_event.emit(f"Connection error: {e}")

        else:
            # Disconnect
            self.event_timer.stop()

            # Disable all events
            for cb in [self.accel_cb, self.button_cb, self.ir_cb, self.battery_cb, self.gpio_cb]:
                cb.setChecked(False)

            if hasattr(self.freewili, "close"):
                self.freewili.close()

            self.freewili = None
            self.connect_btn.setText("Connect")
            self.signals.raw_event.emit("Disconnected from FreeWili")

    def update_event_subscriptions(self):
        """Update event subscriptions based on checkboxes."""
        if self.freewili is None:
            return

        try:
            # Update accelerometer events
            if self.accel_cb.isChecked():
                result = self.freewili.enable_accel_events(True, 100)
                if result.is_err():
                    self.signals.raw_event.emit(f"Error enabling accel: {result.unwrap_err()}")
            else:
                result = self.freewili.enable_accel_events(False)
                if result.is_err():
                    self.signals.raw_event.emit(f"Error disabling accel: {result.unwrap_err()}")

            # Update button events (assuming these methods exist)
            if hasattr(self.freewili, "enable_button_events"):
                if self.button_cb.isChecked():
                    result = self.freewili.enable_button_events(True, 100)
                    if result.is_err():
                        self.signals.raw_event.emit(f"Error enabling buttons: {result.unwrap_err()}")
                else:
                    result = self.freewili.enable_button_events(False)
                    if result.is_err():
                        self.signals.raw_event.emit(f"Error disabling buttons: {result.unwrap_err()}")

            # Update IR events
            if hasattr(self.freewili, "enable_ir_events"):
                if self.ir_cb.isChecked():
                    result = self.freewili.enable_ir_events(True)
                    if result.is_err():
                        self.signals.raw_event.emit(f"Error enabling IR: {result.unwrap_err()}")
                else:
                    result = self.freewili.enable_ir_events(False)
                    if result.is_err():
                        self.signals.raw_event.emit(f"Error disabling IR: {result.unwrap_err()}")

            # Update battery events
            if hasattr(self.freewili, "enable_battery_events"):
                if self.battery_cb.isChecked():
                    result = self.freewili.enable_battery_events(True)
                    if result.is_err():
                        self.signals.raw_event.emit(f"Error enabling battery: {result.unwrap_err()}")
                else:
                    result = self.freewili.enable_battery_events(False)
                    if result.is_err():
                        self.signals.raw_event.emit(f"Error disabling battery: {result.unwrap_err()}")

            # Update GPIO events
            if hasattr(self.freewili, "enable_gpio_events"):
                if self.gpio_cb.isChecked():
                    result = self.freewili.enable_gpio_events(True)
                    if result.is_err():
                        self.signals.raw_event.emit(f"Error enabling GPIO: {result.unwrap_err()}")
                else:
                    result = self.freewili.enable_gpio_events(False)
                    if result.is_err():
                        self.signals.raw_event.emit(f"Error disabling GPIO: {result.unwrap_err()}")

        except Exception as e:
            self.signals.raw_event.emit(f"Error updating subscriptions: {e}")

    def process_events(self):
        """Process FreeWili events (called by timer)."""
        if self.freewili:
            try:
                self.freewili.process_events()
            except Exception as e:
                self.signals.raw_event.emit(f"Event processing error: {e}")

    def closeEvent(self, event):  # noqa: N802
        """Handle window close event."""
        if self.freewili:
            self.toggle_connection()  # Disconnect cleanly
        event.accept()


def main():
    """Main function to run the GUI application."""
    app = QApplication(sys.argv)

    # Set application properties
    app.setApplicationName("FreeWili Events Monitor")
    app.setApplicationVersion("1.0")

    # Create and show the main window
    window = FreeWiliEventsGUI()
    window.show()

    # Run the application
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



"""Tests for freewili.cli_convert module."""

from unittest.mock import patch

from result import Err, Ok

from freewili.cli_convert import main


class TestMain:
    """Test cases for the main function."""

    def test_main_successful_conversion(self) -> None:
        """Test successful image conversion via CLI."""
        test_args = ["fwi-convert", "-i", "input.png", "-o", "output.fwi"]

        with (
            patch("sys.argv", test_args),
            patch("freewili.image.convert", return_value=Ok("Conversion successful")) as mock_convert,
            patch("builtins.print") as mock_print,
        ):
            main()

            mock_convert.assert_called_once_with("input.png", "output.fwi")
            mock_print.assert_called_once_with("Conversion successful")

    def test_main_conversion_failure(self) -> None:
        """Test handling of conversion failure via CLI."""
        test_args = ["fwi-convert", "-i", "input.png", "-o", "output.fwi"]

        with patch("sys.argv", test_args), patch("freewili.image.convert", return_value=Err("Conversion failed")):
            try:
                main()
                raise AssertionError("Should have raised SystemExit")
            except SystemExit as e:
                assert e.code == 1  # Error exit code

    def test_main_missing_input_argument(self) -> None:
        """Test error handling when input argument is missing."""
        test_args = ["fwi-convert", "-o", "output.fwi"]

        with patch("sys.argv", test_args), patch("sys.stderr"):  # Suppress argparse error output
            try:
                main()
                raise AssertionError("Should have raised SystemExit")
            except SystemExit as e:
                assert e.code == 2  # argparse error code

    def test_main_missing_output_argument(self) -> None:
        """Test error handling when output argument is missing."""
        test_args = ["fwi-convert", "-i", "input.png"]

        with patch("sys.argv", test_args), patch("sys.stderr"):  # Suppress argparse error output
            try:
                main()
                raise AssertionError("Should have raised SystemExit")
            except SystemExit as e:
                assert e.code == 2  # argparse error code

    def test_main_version_argument(self) -> None:
        """Test version argument displays version information."""
        test_args = ["fwi-convert", "--version"]

        with (
            patch("sys.argv", test_args),
            patch("importlib.metadata.version", return_value="1.0.0"),
            patch("sys.stdout"),
        ):  # Suppress version output
            try:
                main()
                raise AssertionError("Should have raised SystemExit")
            except SystemExit as e:
                assert e.code == 0  # Normal exit for version display

    def test_main_with_pathlib_paths(self) -> None:
        """Test that string arguments are passed correctly to convert function."""
        test_args = ["fwi-convert", "-i", "/path/to/input.png", "-o", "/path/to/output.fwi"]

        with (
            patch("sys.argv", test_args),
            patch("freewili.image.convert", return_value=Ok("Success")) as mock_convert,
            patch("builtins.print"),
        ):
            main()

            # Verify arguments are passed as strings (argparse provides strings)
            mock_convert.assert_called_once_with("/path/to/input.png", "/path/to/output.fwi")



"""Example script to play audio on the FreeWili."""

import time

from freewili import FreeWili

# https://docs.freewili.com/downloads/FwROMAudioAssets.pdf
AUDIO_NAMES = (
    "1",
    "10",
    "2",
    "3",
    "4",
    "42",
    "5",
    "6",
    "7",
    "8",
    "9",
    "About",
    "AllYourBases",
    "beephigh",
    "ChooseYourDestiny",
    "click",
    "daisy",
    "DaveICantDoThat",
    "DefCon",
    "Dominating",
    "DoOrDoNot",
    "DoubleKill",
    "EndOfLine",
    "Engage",
    "Excellent",
    "ExcuseMe",
    "FearIsTheMindKiller",
    "FreeWilli",
    "GodLike",
    "GoodJob",
    "GPIO",
    "GUI",
    "GUISettings",
    "GUITerminal",
    "HackMe",
    "HangTough",
    "HaveACupcake",
    "HolyShit",
    "I2C",
    "IllBeBack",
    "ImThinking",
    "Infinity",
    "Invalid",
    "IR",
    "ItsTime",
    "KillingSpree",
    "LudacrisKill",
    "MegaKill",
    "Minus",
    "MonsterKill",
    "MultiKill",
    "No",
    "NoWay",
    "NoWayWithLaugh",
    "Orca",
    "palebluedot",
    "PinOut",
    "PlansWithinPlans",
    "Plus",
    "Point",
    "PorqueNo",
    "Radios",
    "Rampage",
    "Revenge",
    "ScanMeHarder",
    "Scripts",
    "Settings",
    "shame",
    "ShutUpWesely",
    "SmileAndBeHappy",
    "Sorry",
    "SoSayWeAll",
    "SPI",
    "Sweet",
    "Tasty",
    "Terminal",
    "ThanksForAllTheFish",
    "tom",
    "UART",
    "UltraKill",
    "UnstopableWind",
    "unstoppable",
    "UserError",
    "VivaLosVegas",
    "WarWarNeverChanges",
    "Welcome",
    "WhaleEver",
    "WhaleThankU",
    "What",
    "WickedSick",
    "Yes",
)

with FreeWili.find_first().expect("Failed to find FreeWili") as fw:
    print(f"Connected to FreeWili {fw}")
    print("type 'list' to see audio assets or 'exit' to quit.")
    while True:
        try:
            user_input = input("Enter to play audio asset (1-{}): ".format(len(AUDIO_NAMES)))
            value = int(user_input) - 1  # Convert to zero-based index
            print(f"\tPlaying: {AUDIO_NAMES[value]}")  # Fixed index here
            fw.play_audio_asset(value).expect("Failed to play audio asset")
            time.sleep(1)  # Give some time for the audio to play, playing is async
        except ValueError:
            if user_input.lower() == "list":
                for i, name in enumerate(AUDIO_NAMES, start=1):
                    print(f"\t{i}: {name}")
                continue
            print("Goodbye!")
            break
    print("Done.")



"""For Interfacing to Free-Wili Devices."""

import pathlib
import platform
import sys
from dataclasses import dataclass
from typing import Any, Callable, List

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import pyfwfinder as fwf
from result import Err, Ok, Result

from freewili.framing import ResponseFrame
from freewili.fw_serial import FreeWiliSerial
from freewili.types import ButtonColor, EventType, FreeWiliProcessorType, IOMenuCommand

# USB Locations:
# first address = FTDI
FTDI_HUB_LOC_INDEX = 2
# second address = Display
DISPLAY_HUB_LOC_INDEX = 1
# third address = Main
MAIN_HUB_LOC_INDEX = 0


class FreeWili:
    """Free-Wili device used to access FTDI and serial functionality."""

    def __init__(self, device: fwf.FreeWiliDevice):
        self.device = device
        self._stay_open = False

        self._main_serial: None | FreeWiliSerial = None
        self._display_serial: None | FreeWiliSerial = None

    def __str__(self) -> str:
        return f"Free-Wili {self.device.serial}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.device.serial}>"

    def __enter__(self) -> Self:
        self.open().expect("Failed to open FreeWili")
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.close()

    def set_event_callback(self, event_cb: None | Callable[[EventType, ResponseFrame, Any], None]) -> None:
        """Set the event callback for the FreeWili.

        Parameters:
        -----------
            event_cb: Callable[[EventType, ResponseFrame, Any], None]
                Callback function to be called when an event occurs.
                The function should accept three arguments: the event type, the response frame,
                and any additional data.
                Set to None to disable the callback.
        """
        if self.main_serial:
            self.main_serial.set_event_callback(event_cb)
        if self.display_serial:
            self.display_serial.set_event_callback(event_cb)

    @property
    def usb_devices(self) -> List[fwf.USBDevice]:
        """Grab all the USB devices attached to the FreeWili."""
        return self.device.usb_devices

    def get_usb_device(self, processor_type: FreeWiliProcessorType) -> None | fwf.USBDevice:
        """Get the USBDevice by ProcessorType.

        Parameters:
        ------------
            processor_type: FreeWiliProcessorType
                Type of processor to get.

        Returns:
        ---------
            None | fwf.USBDevice:
                fwf.USBDevice if successful, None otherwise.

        Raises:
        -------
            None
        """
        match processor_type:
            case FreeWiliProcessorType.Main:
                devs = self.device.get_usb_devices(fwf.USBDeviceType.SerialMain)
                if devs:
                    return devs[0]
            case FreeWiliProcessorType.Display:
                devs = self.device.get_usb_devices(fwf.USBDeviceType.SerialDisplay)
                if devs:
                    return devs[0]
            case FreeWiliProcessorType.FTDI:
                devs = self.device.get_usb_devices(fwf.USBDeviceType.FTDI)
                if devs:
                    return devs[0]
            case FreeWiliProcessorType.ESP32:
                devs = self.device.get_usb_devices(fwf.USBDeviceType.ESP32)
                if devs:
                    return devs[0]
        # Get the processor by location
        # TODO: MAC will be broke here
        if processor_type in (FreeWiliProcessorType.Main, FreeWiliProcessorType.Display):
            for usb_device in self.usb_devices:
                if (
                    processor_type == FreeWiliProcessorType.Main
                    # and usb_device.kind == fwf.USBDeviceType.MassStorage
                    and usb_device.location == 1
                ):
                    return usb_device
                if (
                    processor_type == FreeWiliProcessorType.Display
                    # and usb_device.kind == fwf.USBDeviceType.MassStorage
                    and usb_device.location == 2
                ):
                    return usb_device
        # Legacy support for older VID/PID of Main/Display firmware.
        if len(self.usb_devices) < 3:
            return None
        match processor_type:
            case FreeWiliProcessorType.Main:
                device = self.usb_devices[0]
                return device
            case FreeWiliProcessorType.Display:
                device = self.usb_devices[1]
                return self.usb_devices[1]
        return None

    @property
    def ftdi(self) -> None | fwf.USBDevice:
        """Get FTDI processor."""
        return self.get_usb_device(FreeWiliProcessorType.FTDI)

    @property
    def main(self) -> None | fwf.USBDevice:
        """Get Main processor."""
        return self.get_usb_device(FreeWiliProcessorType.Main)

    @property
    def display(self) -> None | fwf.USBDevice:
        """Get Display processor."""
        return self.get_usb_device(FreeWiliProcessorType.Display)

    @property
    def esp32(self) -> None | fwf.USBDevice:
        """Get Display processor."""
        return self.get_usb_device(FreeWiliProcessorType.ESP32)

    @property
    def main_serial(self) -> None | FreeWiliSerial:
        """Get Main FreeWiliSerial.

        Arguments:
        ----------
            None

        Returns:
        ---------
            None | FreeWiliSerial:
                FreeWiliSerial on success, None otherwise.
        """
        if not self._main_serial and self.main and self.main.port:
            self._main_serial = FreeWiliSerial(self.main.port, self._stay_open, "Main: " + str(self))
        if self._main_serial:
            self._main_serial.stay_open = self._stay_open
        return self._main_serial

    @property
    def display_serial(self) -> None | FreeWiliSerial:
        """Get Display FreeWiliSerial.

        Arguments:
        ----------
            None

        Returns:
        ---------
            None | FreeWiliSerial:
                FreeWiliSerial on success, None otherwise.
        """
        if not self._display_serial and self.display and self.display.port:
            self._display_serial = FreeWiliSerial(self.display.port, self._stay_open, "Display: " + str(self))
        if self._display_serial:
            self._display_serial.stay_open = self._stay_open
        return self._display_serial

    def get_serial_from(self, processor_type: FreeWiliProcessorType) -> Result[FreeWiliSerial, str]:
        """Get FreeWiliSerial from processor type.

        Arguments:
        ----------
            processor_type: FreeWiliProcessorType
                Processor type to get serial port for.

        Returns:
        ---------
            Result[FreeWiliSerial, str]:
                Ok(FreeWiliSerial) on success, Err(str) otherwise.
        """
        match processor_type:
            case FreeWiliProcessorType.Main:
                if self.main_serial:
                    return Ok(self.main_serial)
                else:
                    return Err("Main serial isn't valid")
            case FreeWiliProcessorType.Display:
                if self.display_serial:
                    return Ok(self.display_serial)
                else:
                    return Err("Display serial isn't valid")
            case _:
                return Err(f"{processor_type} isn't a valid FreeWiliSerial type")

    @property
    def stay_open(self) -> bool:
        """Keep serial port open, if True.

        Returns:
            bool
        """
        return self._stay_open

    @stay_open.setter
    def stay_open(self, value: bool) -> None:
        self._stay_open = value

    def open(self, block: bool = True, timeout_sec: float = 6.0) -> Result[None, str]:
        """Open the serial port. Use in conjunction with stay_open.

        Arguments:
        ----------
            block: bool:
                If True, block until the serial port is opened.
            timeout_sec: float:
                number of seconds to wait when blocking.

        Returns:
        ---------
            Result[None, str]:
                Ok(None) if successful, Err(str) otherwise.
        """
        if self.main_serial:
            result = self.main_serial.open(block, timeout_sec)
            if result.is_err():
                return Err(result.err())
        if self.display_serial:
            result = self.display_serial.open(block, timeout_sec)
            if result.is_err():
                return Err(result.err())
        return Ok(None)

    def close(self, restore_menu: bool = True) -> None:
        """Close the serial port. Use in conjunction with stay_open.

        Arguments:
        ----------
            restore_menu: bool
                Restore the menu functionality before close

        Returns:
        ---------
            None
        """
        if self.main_serial:
            self.main_serial.close()
        if self.display_serial:
            self.display_serial.close()

    @classmethod
    def find_first(cls) -> Result[Self, str]:
        """Find first Free-Wili device attached to the host.

        Parameters:
        ------------
            None

        Returns:
        ---------
            Result[FreeWili, str]:
                Ok(FreeWili) if successful, Err(str) otherwise.

        Raises:
        -------
            None
        """
        try:
            devices = cls.find_all()
            if not devices:
                return Err("No FreeWili devices found!")
            return Ok(devices[0])
        except Exception as ex:
            return Err(str(ex))

    @classmethod
    def find_all(cls) -> tuple[Self, ...]:
        """Find all Free-Wili devices attached to the host.

        Parameters:
        ------------
            None

        Returns:
        ---------
            tuple[FreeWili, ...]:
                Tuple of FreeWili devices.

        Raises:
        -------
            None
        """
        found_devices: List[fwf.FreeWiliDevice] = fwf.find_all()
        fw_devices: list[Self] = []
        for device in found_devices:
            fw_devices.append(cls(device))
        return tuple(fw_devices)

    def send_file(
        self,
        source_file: str | pathlib.Path,
        target_name: None | str = None,
        processor: None | FreeWiliProcessorType = None,
        event_cb: Callable | None = None,
        chunk_size: int = 0,
    ) -> Result[str, str]:
        """Send a file to the FreeWili.

        Arguments:
        ----------
            source_file: pathlib.Path
                Path to the file to be sent.
            target_name: None | str
                Name of the file in the FreeWili. If None, will be determined automatically based on the filename.
            processor: None | FreeWiliProcessorType
                Processor to upload the file to. If None, will be determined automatically based on the filename.
            event_cb: Callable | None
                event callback function. Takes one argument of a string.
                    def user_callback(msg: str) -> None
            chunk_size: int
                Size of the chunks to send in bytes. Typically this should be left at the default value.

        Returns:
        ---------
            Result[str, str]:
                Returns Ok(str) if the command was sent successfully, Err(str) if not.
        """
        try:
            # Auto assign values that are None
            if not target_name:
                target_name = FileMap.from_fname(str(source_file)).to_path(str(source_file))
            if not processor:
                processor = FileMap.from_fname(str(source_file)).processor
        except ValueError as ex:
            return Err(str(ex))
        assert target_name is not None
        assert processor is not None

        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.send_file(source_file, target_name, event_cb, chunk_size)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def get_file(
        self,
        source_file: str,
        destination_path: str | pathlib.Path,
        processor: FreeWiliProcessorType | None = None,
        event_cb: Callable | None = None,
    ) -> Result[str, str]:
        """Send a file to the FreeWili.

        Arguments:
        ----------
            source_file: pathlib.Path
                Path to the file to be sent.
            destination_path: pathlib.Path
                file path to save on the PC
            processor: None | FreeWiliProcessorType
                Processor to upload the file to. If None, will be determined automatically based on the filename.
            event_cb: Callable | None
                event callback function. Takes one argument of a string.
                    def user_callback(msg: str) -> None

        Returns:
        ---------
            Result[str, str]:
                Returns Ok(str) if the command was sent successfully, Err(str) if not.
        """
        try:
            # Auto assign values that are None
            if not processor:
                processor = FileMap.from_fname(str(source_file)).processor
        except ValueError as ex:
            return Err(str(ex))
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.get_file(source_file, destination_path, event_cb)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def run_script(
        self, file_name: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[str, str]:
        """Run a script on the FreeWili.

        Arguments:
        ----------
            file_name: str
                Name of the file in the FreeWili. 8.3 filename limit exists as of V12
            processor: FreeWiliProcessorType
                Processor to upload the file to.

        Returns:
        ---------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.run_script(file_name)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def get_io(self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main) -> Result[tuple[int, ...], str]:
        """Get all the IO values.

        Parameters:
        ------------
            processor: FreeWiliProcessorType
                Processor to set IO on.

        Returns:
        ---------
            Result[tuple[int], str]:
                Ok(tuple[int]) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.get_io()
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def toggle_high_speed_io(
        self: Self,
        enable: bool,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
    ) -> Result[str, str]:
        """Toggle the high-speed Bidirectional IO.

        Parameters:
        ------------
            enable: bool
                Whether to enable or disable high-speed IO.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.toggle_high_speed_io(enable)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def set_io(
        self: Self,
        io: int,
        menu_cmd: IOMenuCommand,
        pwm_freq: None | int = None,
        pwm_duty: None | int = None,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
    ) -> Result[str, str]:
        """Set the state of an IO pin to high or low.

        Parameters:
        ------------
            io : int
                The number of the IO pin to set.
            menu_cmd : IOMenuCommand
                Whether to set the pin to high, low, toggle, or pwm.
            pwm_freq: None | int
                PWM frequency in Hertz
            pwm_duty: None | int
                PWM Duty cycle (0-100)
            processor: FreeWiliProcessorType
                Processor to set IO on.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.set_io(io, menu_cmd, pwm_freq, pwm_duty)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def set_board_leds(
        self: Self,
        io: int,
        red: int,
        green: int,
        blue: int,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
    ) -> Result[ResponseFrame, str]:
        """Set the GUI RGB LEDs.

        Parameters:
        ------------
            io : int
                The number of the IO pin to set.
            red : int
                Red Color 0-255
            green : int
                Green Color 0-255
            blue : int
                Blue Color 0-255
            processor: FreeWiliProcessorType
                Processor to set LEDs on.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.set_board_leds(io, red, green, blue)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def read_i2c(
        self, address: int, register: int, data_size: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[bytes, str]:
        """Write I2C data.

        Parameters:
        ------------
            address : int
                The address to write to.
            register : int
                The register to write to.
            data_size : int
                The number of bytes to read.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.read_i2c(address, register, data_size)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def write_i2c(
        self, address: int, register: int, data: bytes, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[str, str]:
        """Write I2C data.

        Parameters:
        -----------
            address : int
                The address to write to.
            register : int
                The register to write to.
            data : bytes
                The data to write.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
                The str is the response from the device, typically "OK".
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.write_i2c(address, register, data)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def poll_i2c(self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main) -> Result[tuple[int, ...], str]:
        """Poll I2C data.

        Parameters:
        ------------
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[tuple[int, ...], str]:
                Ok(tuple[int, ...]) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.poll_i2c()
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def show_gui_image(
        self, fwi_path: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[ResponseFrame, str]:
        """Show a fwi image on the display.

        Arguments:
        ----------
            fwi_path: str
                path to the fwi image
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.show_gui_image(fwi_path)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def show_text_display(
        self, text: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[ResponseFrame, str]:
        """Show text on the display.

        Arguments:
        ----------
            text: str
                text to display on screen.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.show_text_display(text)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def read_all_buttons(
        self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[dict[ButtonColor, bool], str]:
        """Read all the buttons.

        Arguments:
        ----------
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[dict[ButtonColor, bool], str]:
                Ok(dict[ButtonColor, bool]) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.read_all_buttons()
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def reset_display(
        self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[ResponseFrame, str]:
        """Reset the display back to the main menu.

        Arguments:
        ----------
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.reset_display()
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def enable_accel_events(
        self,
        enable: bool,
        interval_ms: int | None = None,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
    ) -> Result[str, str]:
        """Enable or disable acceleration events.

        Arguments:
        ----------
            enable: bool
                Whether to enable or disable acceleration events.
            interval_ms: int | None
                The interval in milliseconds for accelerometer events. If None, the default value will be used.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.enable_accel_events(enable, interval_ms)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def enable_gpio_events(
        self,
        enable: bool,
        interval_ms: int | None = None,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
    ) -> Result[str, str]:
        """Enable or disable GPIO events.

        Arguments:
        ----------
            enable: bool
                Whether to enable or disable GPIO events.
            interval_ms: int | None
                The interval in milliseconds for GPIO events. If None, the default value will be used.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.enable_gpio_events(enable, interval_ms)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def enable_button_events(
        self,
        enable: bool,
        interval_ms: int | None = None,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
    ) -> Result[str, str]:
        """Enable or disable acceleration events.

        Arguments:
        ----------
            enable: bool
                Whether to enable or disable acceleration events.
            interval_ms: int | None
                The interval in milliseconds for accelerometer events. If None, the default value will be used.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.enable_button_events(enable, interval_ms)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def enable_ir_events(
        self,
        enable: bool,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
    ) -> Result[str, str]:
        """Enable or disable infrared events.

        Arguments:
        ----------
            enable: bool
                Whether to enable or disable infrared events.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.enable_ir_events(enable)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def enable_battery_events(
        self,
        enable: bool,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
    ) -> Result[str, str]:
        """Enable or disable battery events.

        Arguments:
        ----------
            enable: bool
                Whether to enable or disable battery events.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.enable_battery_events(enable)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def enable_radio_events(
        self,
        enable: bool,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
    ) -> Result[str, str]:
        """Enable or disable radio events on currently selected radio.

        Arguments:
        ----------
            enable: bool
                Whether to enable or disable radio events.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.enable_radio_events(enable)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def enable_uart_events(
        self,
        enable: bool,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
    ) -> Result[str, str]:
        """Enable or disable UART events.

        Arguments:
        ----------
            enable: bool
                Whether to enable or disable UART events.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.enable_uart_events(enable)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def enable_audio_events(
        self,
        enable: bool,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
    ) -> Result[str, str]:
        """Enable or disable audio events.

        Arguments:
        ----------
            enable: bool
                Whether to enable or disable audio events.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[ResponseFrame, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.enable_audio_events(enable)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def process_events(self) -> None:
        """Process any events that have been received.

        This is typically called in a loop to process events as they come in.
        """
        if self.main_serial:
            self.main_serial.process_events()
        if self.display_serial:
            self.display_serial.process_events()

    def select_radio(
        self, radio_index: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[str, str]:
        """Select the radio to use for events.

        Arguments:
        ----------
            radio_index: int
                Index of the radio to select. 1 or 2 typically.

            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.select_radio(radio_index)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def set_radio_event_rssi_threshold(
        self, rssi: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[str, str]:
        """Set the RSSI threshold for radio events.

        Arguments:
        ----------
            rssi: int
                RSSI threshold value to set.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.set_radio_event_rssi_threshold(rssi)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def set_radio_event_sample_window(
        self, sample_window_ms: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[str, str]:
        """Set the sample window (ms) for the specified radio.

        Arguments:
        ----------
            sample_window_ms: int
                Sample window value to set.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.set_radio_event_sample_window(sample_window_ms)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def transmit_radio_subfile(
        self, sub_fname: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[str, str]:
        """Transmit a subfile to the specified radio.

        Arguments:
        ----------
            sub_fname: str
                Name of the subfile to transmit. This should be the filename with the extension.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.transmit_radio_subfile(sub_fname)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def write_radio(
        self, data: bytes, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[str, str]:
        """Write radio data.

        Arguments:
        ----------
            data: bytes
                Data to transmit.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.write_radio(data)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def write_uart(
        self, data: bytes | str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[str, str]:
        """Write uart data.

        Arguments:
        ----------
            data: bytes
                Data to transmit.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.write_uart(data)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def send_ir(
        self, data: bytes, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[str, str]:
        """Send IR data.

        Notes: v54 firmware uses NEC format and is converted to an 32-bit integer.

        Parameters:
        ----------
            data : bytes
                The data to send. The first 4 bytes are used as the command.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[bytes, str]:
                Ok(bytes) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.send_ir(data)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def play_audio_file(
        self, file_name: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[str, str]:
        """Play an audio file on the FreeWili.

        Arguments:
        ----------
            file_name: str
                Name of the file in the FreeWili. 8.3 filename limit exists as of V12
            processor: FreeWiliProcessorType
                Processor to upload the file to.

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.play_audio_file(file_name)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def play_audio_asset(
        self, asset_value: str | int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[str, str]:
        """Play an audio asset on the FreeWili.

        Arguments:
        ----------
            asset_value: str | int
                The asset value to play.
            processor: FreeWiliProcessorType
                Processor to upload the file to.

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.play_audio_asset(asset_value)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def play_audio_number_as_speech(
        self, value: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[str, str]:
        """Play an audio number as speech on the FreeWili.

        Arguments:
        ----------
            value: int
                The audio number to play.
            processor: FreeWiliProcessorType
                Processor to upload the file to.

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.play_audio_number_as_speech(value)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def play_audio_tone(
        self,
        frequency_hz: int,
        duration_sec: float,
        amplitude: float,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
    ) -> Result[str, str]:
        """Play an audio tone on the FreeWili.

        Arguments:
        ----------
            frequency_hz: int
                The frequency of the tone in Hertz.
            duration_sec: float
                The duration of the tone in seconds.
            amplitude: float
                The amplitude of the tone (0.0 to 1.0).
            processor: FreeWiliProcessorType
                Processor to upload the file to.

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.play_audio_tone(frequency_hz, duration_sec, amplitude)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def record_audio(
        self, file_name: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[str, str]:
        """Record audio on the FreeWili.

        Arguments:
        ----------
            file_name: str
                Name of the file in the FreeWili. (ie. "/sounds/test.wav")
            processor: FreeWiliProcessorType
                Processor to upload the file to.

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.record_audio(file_name)
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")

    def reset_to_uf2_bootloader(self, processor: FreeWiliProcessorType) -> Result[None, str]:
        """Reset the FreeWili to the uf2 bootloader.

        Arguments:
        ----------
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        ---------
            Result[None, str]:
                Returns Ok(None) if the command was sent successfully, Err(str) if not.
        """
        match self.get_serial_from(processor):
            case Ok(serial):
                return serial.reset_to_uf2_bootloader()
            case Err(msg):
                return Err(msg)
            case _:
                raise RuntimeError("Missing case statement")


@dataclass(frozen=True)
class FileMap:
    """Map file extension to processor type and location."""

    # file extension type (ie. .fwi)
    extension: str
    # processor the file should live on
    processor: FreeWiliProcessorType
    # directory the file type
    directory: str
    # description of the file type
    description: str

    @classmethod
    def from_ext(cls, ext: str) -> Self:
        """Creates a FileMap from a file extension.

        Parameters:
        ------------
            ext: str
                File extension (ie. ".wasm"). Not case sensitive.

        Returns:
        ---------
            FileMap

        Raises:
        -------
            ValueError:
                If the extension isn't known.
        """
        ext = ext.lstrip(".").lower()
        mappings = {
            "wasm": (FreeWiliProcessorType.Main, "/scripts", "WASM binary"),
            "wsm": (FreeWiliProcessorType.Main, "/scripts", "WASM binary"),
            "zio": (FreeWiliProcessorType.Main, "/scripts", "ZoomIO script file"),
            "bin": (FreeWiliProcessorType.Main, "/fpga", "FPGA bin file"),
            "sub": (FreeWiliProcessorType.Main, "/radio", "Radio file"),
            "fwi": (FreeWiliProcessorType.Display, "/images", "Image file"),
            "wav": (FreeWiliProcessorType.Display, "/sounds", "Audio file"),
        }
        if ext not in mappings:
            raise ValueError(f"Extension '{ext}' is not a known FreeWili file type")
        return cls(ext, *mappings[ext])

    @classmethod
    def from_fname(cls, file_name: str) -> Self:
        """Creates a FileMap from a file path.

        Parameters:
        ------------
            file_name: str
                File name (ie. "myfile.wasm"). Not case sensitive. Can contain paths.

        Returns:
        ---------
            FileMap

        Raises:
        -------
            ValueError:
                If the extension isn't known.
        """
        fpath = pathlib.Path(file_name)
        return cls.from_ext(fpath.suffix)

    def to_path(self, file_name: str) -> str:
        """Creates a file path from the file_name to upload to the FreeWili.

        Parameters:
        ------------
            file_name: str
                File name (ie. "myfile.wasm"). Not case sensitive. Can contain paths.

        Returns:
        ---------
            str
                Full file path intended to be uploaded to a FreeWili

        Raises:
        -------
            ValueError:
                If the extension isn't known.
        """
        fpath = pathlib.Path(file_name)
        fpath_str = str(pathlib.Path(self.directory) / fpath.name)
        if platform.system().lower() == "windows":
            fpath_str = fpath_str.replace("\\", "/")
        return fpath_str


if __name__ == "__main__":
    devices = FreeWili.find_all()
    print(f"Found {len(devices)} Free-Wili(s):")
    for i, dev in enumerate(devices):
        print(f"{i}. {dev}")
        ftdi = dev.ftdi
        main = dev.main
        display = dev.display
        print("\tFTDI:   ", ftdi)  # type: ignore
        print("\tMain:   ", main)  # type: ignore
        print("\tDisplay:", display)  # type: ignore



"""Test SerialPort Module."""

import time
import unittest
from typing import Any
from unittest.mock import MagicMock, patch

from freewili.serialport import SerialPort  # Adjust this import as needed


class TestSerialPort(unittest.TestCase):
    """Test Serial Port Class."""

    @patch("freewili.serialport.Serial")  # Patch Serial from pyserial
    def test_send_queue_gets_called(self, mock_serial_class: Any) -> None:
        """Test send queue."""
        # Create a mock Serial instance
        mock_serial_instance = MagicMock()
        mock_serial_instance.in_waiting = 0
        mock_serial_class.return_value = mock_serial_instance
        mock_serial_instance.write.return_value = len(b"test\n")
        mock_serial_instance.read.return_value = b""

        # Create the SerialReader instance and start the thread
        serial_port: SerialPort = SerialPort(port="COM1")
        serial_port.open()

        # Send data to the thread
        test_data = "test"
        serial_port.send(test_data)

        # Wait until write() is called or timeout
        for _ in range(50):  # Wait up to ~0.5s
            if mock_serial_instance.write.called:
                break
            time.sleep(0.01)

        # Close the thread gracefully
        serial_port.close()

        # Check that write() was called with the correct data
        expected_data = b"test\n"
        mock_serial_instance.write.assert_called_with(expected_data)

        # Optionally check send_queue is empty now
        self.assertTrue(serial_port.send_queue.empty())


if __name__ == "__main__":
    import pytest

    pytest.main(
        args=[
            __file__,
            "--verbose",
        ]
    )



"""Set Board RGB on a Free-Wili."""

import time

from freewili import FreeWili

# find a FreeWili device and open it
device = FreeWili.find_first().expect("Failed to find a FreeWili")
device.open().expect("Failed to open")

# Turn the LEDs on
for led_num in range(7):
    print(device.set_board_leds(led_num, 10, 10, led_num * 2).expect("Failed to set LED"))
# Wait so we can see them
time.sleep(3)
# Turn the LEDS off
for led_num in reversed(range(7)):
    print(device.set_board_leds(led_num, 0, 0, 0).expect("Failed to set LED"))

device.close()
